"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var assert = require("assert");
var TaskStatus_1 = require("./TaskStatus");
/**
 * This is the abstract base class of all behavior tree tasks.
 * @param T The type of the blackboard object that tasks use to read or modify game state.
 */
var BaseTask = /** @class */ (function () {
    function BaseTask() {
        this.status = TaskStatus_1.TaskStatus.Invalid;
    }
    BaseTask.prototype.getStatus = function () {
        return this.status;
    };
    BaseTask.prototype.getObject = function () {
        if (this.tree === undefined) {
            throw new Error('This task has never run.');
        }
        return this.tree.getObject();
    };
    BaseTask.prototype.setGuard = function (guard) {
        this.guard = guard;
        return this;
    };
    BaseTask.prototype.setControl = function (control) {
        this.control = control;
        this.tree = control.tree;
    };
    BaseTask.prototype.checkGuard = function (control) {
        if (this.guard === undefined) {
            // No guard to check.
            return true;
        }
        // Check the guard of the guard recursively.
        if (!this.guard.checkGuard(control)) {
            return false;
        }
        if (control.tree === undefined) {
            throw new Error('Invalid state.');
        }
        // Use the tree's guard evaluator task to check the guard of this task.
        this.guard.setControl(control.tree.guardEvaluator);
        this.guard.start();
        this.guard.run();
        var status = this.guard.getStatus();
        if (status === TaskStatus_1.TaskStatus.Succeeded) {
            return true;
        }
        if (status === TaskStatus_1.TaskStatus.Failed) {
            return false;
        }
        throw new Error('Illegal guard status. Guard must either succeed or fail in one step.');
    };
    BaseTask.prototype.start = function () { };
    BaseTask.prototype.end = function () { };
    BaseTask.prototype.success = function () {
        this.status = TaskStatus_1.TaskStatus.Succeeded;
        this.end();
        this.control && this.control.childSuccess(this);
    };
    BaseTask.prototype.fail = function () {
        this.status = TaskStatus_1.TaskStatus.Failed;
        this.end();
        this.control && this.control.childFail(this);
    };
    BaseTask.prototype.running = function () {
        this.status = TaskStatus_1.TaskStatus.Running;
        this.control && this.control.childRunning(this, this);
    };
    BaseTask.prototype.cancel = function () {
        assert(this.getStatus() === TaskStatus_1.TaskStatus.Running);
        this.status = TaskStatus_1.TaskStatus.Canceled;
        this.end();
    };
    /** Terminates the running children of this task starting from the specified inex up to the end. */
    BaseTask.prototype.cancelRunningChildren = function (startIndex) {
        for (var i = startIndex, n = this.getChildCount(); i < n; ++i) {
            var child = this.getChild(i);
            if (child.getStatus() === TaskStatus_1.TaskStatus.Running) {
                child.cancel();
            }
        }
    };
    /** Resets this task to make it restart from scratch on next run. */
    BaseTask.prototype.resetTask = function () {
        if (this.status === TaskStatus_1.TaskStatus.Running) {
            this.cancel();
        }
        for (var i = 0, n = this.getChildCount(); i < n; ++i) {
            var child = this.getChild(i);
            child.resetTask();
        }
        this.status = TaskStatus_1.TaskStatus.Invalid;
        this.control = undefined;
        this.tree = undefined;
    };
    BaseTask.prototype.reset = function () {
        this.status = TaskStatus_1.TaskStatus.Invalid;
        this.guard = undefined;
        this.control = undefined;
        this.tree = undefined;
    };
    return BaseTask;
}());
exports.BaseTask = BaseTask;
